from reconfigure.items.bound import BoundData


class PasswdData (BoundData):
    pass


class UserData (BoundData):
    fields = ['name', 'password', 'uid', 'gid', 'comment', 'home', 'shell']


PasswdData.bind_collection('users', item_class=UserData)
for i in range(0, len(UserData.fields)):
    path = lambda i: lambda x: x.children[i]
    UserData.bind_property('value', UserData.fields[i], path=path(i))



class BaseBuilder (object):
    """
    A base class for builders
    """

    def build(self, tree):
        """
        :param tree: :class:`reconfigure.nodes.Node` tree
        :returns: Data tree
        """

    def unbuild(self, tree):
        """
        :param tree: Data tree
        :returns: :class:`reconfigure.nodes.Node` tree
        """



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData
from reconfigure.items.util import tf_getter, tf_setter


class SupervisorData (BoundData):
    pass


class ProgramData (BoundData):
    fields = ['command', 'autostart', 'autorestart', 'startsecs', 'startretries', \
        'user', 'directory', 'umask', 'environment', 'stopasgroup', 'killasgroup']

    def template(self):
        return Node('program:new',
            PropertyNode('command', 'false'),
        )


SupervisorData.bind_collection('programs', item_class=ProgramData, selector=lambda x: x.name.startswith('program:'))
ProgramData.bind_name('name', getter=lambda x: x[8:], setter=lambda x: 'program:%s' % x)
ProgramData.bind_attribute('comment', 'comment')
for i in range(0, len(ProgramData.fields)):
    ProgramData.bind_property(ProgramData.fields[i], ProgramData.fields[i], default_remove=[None, ''])
ProgramData.bind_property('stopasgroup', 'stopasgroup', default_remove=[False], getter=tf_getter, setter=tf_setter)
ProgramData.bind_property('killasgroup', 'killasgroup', default_remove=[False], getter=tf_getter, setter=tf_setter)



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import SSVParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.passwd import PasswdData


class PasswdConfig (Reconfig):
    """
    ``/etc/passwd``
    """

    def __init__(self, **kwargs):
        k = {
            'parser': SSVParser(separator=':'),
            'builder': BoundBuilder(PasswdData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



class BaseParser (object):  # pragma: no cover
    """
    A base parser class
    """

    def parse(self, content):
        """
        :param content: string config content
        :returns: a :class:`reconfigure.nodes.Node` tree
        """
        return None

    def stringify(self, tree):
        """
        :param tree: a :class:`reconfigure.nodes.Node` tree
        :returns: string config content
        """
        return None



from reconfigure.includers.auto import AutoIncluder
from reconfigure.nodes import Node, PropertyNode


class SupervisorIncluder (AutoIncluder):
    def is_include(self, node):
        if node.name == 'include':
            return node.get('files').value

    def remove_include(self, node):
        return Node('include', children=[PropertyNode('files', node.files)])



from reconfigure.parsers.base import BaseParser
from reconfigure.parsers.bind9 import BIND9Parser
from reconfigure.parsers.exports import ExportsParser
from reconfigure.parsers.ini import IniFileParser
from reconfigure.parsers.iptables import IPTablesParser
from reconfigure.parsers.jsonparser import JsonParser
from reconfigure.parsers.nginx import NginxParser
from reconfigure.parsers.nsd import NSDParser
from reconfigure.parsers.shell import ShellParser
from reconfigure.parsers.ssv import SSVParser
from reconfigure.parsers.squid import SquidParser
from reconfigure.parsers.crontab import CrontabParser

__all__ = [
    'BaseParser',
    'BIND9Parser',
    'CrontabParser',
    'ExportsParser',
    'IniFileParser',
    'IPTablesParser',
    'JsonParser',
    'NginxParser',
    'NSDParser',
    'ShellParser',
    'SSVParser',
    'SquidParser',
]



from reconfigure.items.bound import BoundData


class GroupsData (BoundData):
    pass


class GroupData (BoundData):
    fields = ['name', 'password', 'gid', 'users']


GroupsData.bind_collection('groups', item_class=GroupData)
for i in range(0, len(GroupData.fields)):
    path = lambda i: lambda x: x.children[i]
    GroupData.bind_property('value', GroupData.fields[i], path=path(i))



from reconfigure.nodes import *
from reconfigure.parsers.nginx import NginxParser


class BIND9Parser (NginxParser):
    """
    A parser for named.conf
    """

    tokens = [
        (r"(acl|key|masters|server|trusted-keys|managed-keys|controls|logging|lwres|options|view|zone|channel|category|listen-on|search|avoid-v4-udp-ports|avoid-v6-udp-ports|blackhole|listen-on|listen-on-v6|allow-recursion|allow-recursion-on|sortlist|topology|rrset-order|dual-stack-servers|disable-algorithms|dns64|forwarders|rrset-order|update-policy|also-notify|allow-notify|rate-limit)\s+?([^\s{}]*\s*)*{", lambda s, t: ('section_start', t)),
        (r"\#.*?\n", lambda s, t: ('comment', t)),
        (r"//.*?\n", lambda s, t: ('comment', t)),
        (r"/\*.*?\*/", lambda s, t: ('comment', t)),
        (r"((([^\s{};#]+)|({\s*([^\s{};#]+;\s*)*}))\s*?)+;", lambda s, t: ('option', t)),
        (r"\s", lambda s, t: 'whitespace'),
        (r"$^", lambda s, t: 'newline'),
        (r"\};", lambda s, t: 'section_end'),
    ]
    token_section_end = '};'



class BaseIncluder (object):  # pragma: no cover
    """
    A base includer class

    :param parser: Parser instance that was used to parse the root config file
    :param content_map: a dict that overrides config content for specific paths
    """

    def __init__(self, parser=None, content_map={}):
        self.parser = parser
        self.content_map = content_map

    def compose(self, origin, tree):
        """
        Should locate the include nodes in the Node tree, replace them with :class:`reconfigure.nodes.IncludeNode`, parse the specified include files and append them to tree, with correct node ``origin`` attributes
        """

    def decompose(self, origin, tree):
        """
        Should detach the included subtrees from the Node tree and return a ``{ origin: content-node-tree }`` dict.
        """



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import NSDParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.nsd import NSDData


class NSDConfig (Reconfig):
    """
    ``NSD DNS server nsd.conf``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': NSDParser(),
            'builder': BoundBuilder(NSDData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser


class IPTablesParser (BaseParser):
    """
    A parser for ``iptables`` configuration as produced by ``iptables-save``
    """

    def parse(self, content):
        content = filter(None, [x.strip() for x in content.splitlines() if not x.startswith('#')])
        root = RootNode()
        cur_table = None
        chains = {}
        for l in content:
            if l.startswith('*'):
                cur_table = Node(l[1:])
                chains = {}
                root.append(cur_table)
            elif l.startswith(':'):
                name = l[1:].split()[0]
                node = Node(name)
                node.set_property('default', l.split()[1])
                chains[name] = node
                cur_table.append(node)
            else:
                comment = None
                if '#' in l:
                    l, comment = l.split('#')
                    comment = comment.strip()
                tokens = l.split()
                if tokens[0] == '-A':
                    tokens.pop(0)
                    node = Node('append')
                    node.comment = comment
                    chain = tokens.pop(0)
                    chains[chain].append(node)
                    while tokens:
                        token = tokens.pop(0)
                        option = Node('option')
                        option.set_property('negative', token == '!')
                        if token == '!':
                            token = tokens.pop(0)
                        option.set_property('name', token.strip('-'))
                        while tokens and not tokens[0].startswith('-') and tokens[0] != '!':
                            option.append(Node('argument', PropertyNode('value', tokens.pop(0))))
                        node.append(option)

        return root

    def stringify(self, tree):
        data = ''
        for table in tree.children:
            data += '*%s\n' % table.name
            for chain in table.children:
                data += ':%s %s [0:0]\n' % (chain.name, chain.get('default').value)
            for chain in table.children:
                for item in chain.children:
                    if item.name == 'append':
                        data += '-A %s %s%s\n' % (
                            chain.name,
                            ' '.join(
                                ('! ' if o.get('negative').value else '') +
                                ('--' if len(o.get('name').value) > 1 else '-') + o.get('name').value + ' ' +
                                ' '.join(a.get('value').value for a in o.children if a.name == 'argument')
                                for o in item.children
                                if o.name == 'option'
                            ),
                            ' # %s' % item.comment if item.comment else ''
                        )
            data += 'COMMIT\n'
        return data



"""
Configs are ready-to-use objects that link together Parsers, Includers and Builders to provide direct conversion between config files and Data tree.
"""

from reconfigure.configs.base import Reconfig
from reconfigure.configs.ajenti import AjentiConfig
from reconfigure.configs.bind9 import BIND9Config
from reconfigure.configs.crontab import CrontabConfig
from reconfigure.configs.ctdb import CTDBConfig, CTDBNodesConfig, CTDBPublicAddressesConfig
from reconfigure.configs.csf import CSFConfig
from reconfigure.configs.dhcpd import DHCPDConfig
from reconfigure.configs.exports import ExportsConfig
from reconfigure.configs.fstab import FSTabConfig
from reconfigure.configs.group import GroupConfig
from reconfigure.configs.hosts import HostsConfig
from reconfigure.configs.iptables import IPTablesConfig
from reconfigure.configs.netatalk import NetatalkConfig
from reconfigure.configs.nsd import NSDConfig
from reconfigure.configs.passwd import PasswdConfig
from reconfigure.configs.resolv import ResolvConfig
from reconfigure.configs.samba import SambaConfig
from reconfigure.configs.squid import SquidConfig
from reconfigure.configs.supervisor import SupervisorConfig


__all__ = [
    'Reconfig',
    'AjentiConfig',
    'BIND9Config',
    'CrontabConfig',
    'CTDBConfig',
    'CTDBNodesConfig',
    'CTDBPublicAddressesConfig',
    'DHCPDConfig',
    'ExportsConfig',
    'FSTabConfig',
    'GroupConfig',
    'HostsConfig',
    'IPTablesConfig',
    'NetatalkConfig',
    'NSDConfig',
    'PasswdConfig',
    'ResolvConfig',
    'SambaConfig',
    'SquidConfig',
    'SupervisorConfig',
]



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class NSDData (BoundData):
    pass


class ZoneData (BoundData):
    def template(self):
        return Node(
            'zone',
            PropertyNode('name', '"example.com"'),
            PropertyNode('file', '"example.com.zone"'),
        )


quote = lambda x: '"%s"' % x
unquote = lambda x: x.strip('"')

NSDData.bind_collection('zones', selector=lambda x: x.name == 'zone', item_class=ZoneData)
ZoneData.bind_property('name', 'name', getter=unquote, setter=quote)
ZoneData.bind_property('zonefile', 'file', getter=unquote, setter=quote)



from reconfigure.includers.auto import AutoIncluder
from reconfigure.nodes import PropertyNode


class BIND9Includer (AutoIncluder):
    def is_include(self, node):
        if isinstance(node, PropertyNode) and node.name == 'include':
            return node.value.strip('"')

    def remove_include(self, node):
        return PropertyNode('include', '"%s"' % node.files)



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import SquidParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.squid import SquidData


class SquidConfig (Reconfig):
    def __init__(self, **kwargs):
        k = {
            'parser': SquidParser(),
            'builder': BoundBuilder(SquidData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser


class ShellParser (BaseParser):
    """
    A parser for shell scripts with variables
    """

    def __init__(self, *args, **kwargs):
        self.comment = '#'
        self.continuation = '\\'
        BaseParser.__init__(self, *args, **kwargs)

    def parse(self, content):
        rawlines = content.splitlines()
        lines = []
        while rawlines:
            l = rawlines.pop(0).strip()
            while self.continuation and rawlines and l.endswith(self.continuation):
                l = l[:-len(self.continuation)]
                l += rawlines.pop(0)
            lines.append(l)

        root = RootNode()
        last_comment = None
        for line in lines:
            line = line.strip()
            if line:
                if line.startswith(self.comment):
                    c = line.strip(self.comment).strip()
                    if last_comment:
                        last_comment += '\n' + c
                    else:
                        last_comment = c
                    continue
                if len(line) == 0:
                    continue

                name, value = line.split('=', 1)
                comment = None
                if '#' in value:
                    value, comment = value.split('#', 1)
                    last_comment = (last_comment or '') + comment.strip()
                node = PropertyNode(name.strip(), value.strip().strip('"'))
                if last_comment:
                    node.comment = last_comment
                    last_comment = None
                root.append(node)
        return root

    def stringify(self, tree):
        r = ''
        for node in tree.children:
            if node.comment and '\n' in node.comment:
                r += '\n' + ''.join('# %s\n' % x for x in node.comment.splitlines())
            r += '%s="%s"' % (node.name, node.value)
            if node.comment and not '\n' in node.comment:
                r += ' # %s' % node.comment
            r += '\n'
        return r



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser
from reconfigure.parsers.iniparse import INIConfig

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class IniFileParser (BaseParser):
    """
    A parser for standard ``.ini`` config files.

    :param sectionless: if ``True``, allows a section-less attributes appear in the beginning of file
    """

    def __init__(self, sectionless=False, nullsection='__default__'):
        self.sectionless = sectionless
        self.nullsection = nullsection

    def _get_comment(self, container):
        c = container.contents[0].comment
        return c.strip() if c else None

    def _set_comment(self, container, comment):
        if comment:
            container.contents[0].comment = comment
            container.contents[0].comment_separator = ';'

    def parse(self, content):
        content = '\n'.join(filter(None, [x.strip() for x in content.splitlines()]))
        if self.sectionless:
            content = '[' + self.nullsection + ']\n' + content
        data = StringIO(content)
        cp = INIConfig(data, optionxformvalue=lambda x: x)

        root = RootNode()
        for section in cp:
            name = section
            if self.sectionless and section == self.nullsection:
                name = None
            section_node = Node(name)
            section_node.comment = self._get_comment(cp[section]._lines[0])
            for option in cp[section]:
                if option in cp[section]._options:
                    node = PropertyNode(option, cp[section][option])
                    node.comment = self._get_comment(cp[section]._options[option])
                    section_node.children.append(node)
            root.children.append(section_node)
        return root

    def stringify(self, tree):
        cp = INIConfig()
        for section in tree.children:
            if self.sectionless and section.name is None:
                sectionname = self.nullsection
            else:
                sectionname = section.name
            cp._new_namespace(sectionname)
            for option in section.children:
                if not isinstance(option, PropertyNode):
                    raise TypeError('Third level nodes should be PropertyNodes')
                cp[sectionname][option.name] = option.value
                if option.comment:
                    self._set_comment(cp[sectionname]._options[option.name], option.comment)
            if hasattr(cp[sectionname], '_lines'):
                self._set_comment(cp[sectionname]._lines[0], section.comment)

        data = str(cp) + '\n'
        if self.sectionless:
            data = data.replace('[' + self.nullsection + ']\n', '')
        return data



from reconfigure.includers.base import BaseIncluder
from reconfigure.includers.auto import AutoIncluder
from reconfigure.includers.bind9 import BIND9Includer
from reconfigure.includers.nginx import NginxIncluder
from reconfigure.includers.supervisor import SupervisorIncluder


__all__ = [
    'BaseIncluder',
    'AutoIncluder',
    'BIND9Includer',
    'NginxIncluder',
    'SupervisorIncluder',
]



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser
import json


class JsonParser (BaseParser):
    """
    A parser for JSON files (using ``json`` module)
    """

    def parse(self, content):
        node = RootNode()
        self.load_node_rec(node, json.loads(content))
        return node

    def load_node_rec(self, node, json):
        for k in sorted(json.keys()):
            v = json[k]
            if isinstance(v, dict):
                child = Node(k)
                node.children.append(child)
                self.load_node_rec(child, v)
            else:
                node.children.append(PropertyNode(k, v))

    def stringify(self, tree):
        return json.dumps(self.save_node_rec(tree), indent=4)

    def save_node_rec(self, node):
        r = {}
        for child in node.children:
            if isinstance(child, PropertyNode):
                r[child.name] = child.value
            else:
                r[child.name] = self.save_node_rec(child)
        return r
